# Auto-generated by asn1ate v.0.6.0 from ts_119_495.asn1
# (last modified on 2023-03-15 09:20:01.431627)

from pyasn1.type import univ, char, namedtype, constraint


def _OID(*components):
    output = []
    for x in tuple(components):
        if isinstance(x, univ.ObjectIdentifier):
            output.extend(list(x))
        else:
            output.append(int(x))

    return univ.ObjectIdentifier(output)


class NCAId(char.UTF8String):
    pass


NCAId.subtypeSpec = constraint.ValueSizeConstraint(1, 256)


class NCAName(char.UTF8String):
    pass


NCAName.subtypeSpec = constraint.ValueSizeConstraint(1, 256)


class RoleOfPspOid(univ.ObjectIdentifier):
    pass


class RoleOfPspName(char.UTF8String):
    pass


RoleOfPspName.subtypeSpec = constraint.ValueSizeConstraint(1, 256)


class RoleOfPSP(univ.Sequence):
    pass


RoleOfPSP.componentType = namedtype.NamedTypes(
    namedtype.NamedType('roleOfPspOid', RoleOfPspOid()),
    namedtype.NamedType('roleOfPspName', RoleOfPspName())
)


class RolesOfPSP(univ.SequenceOf):
    pass


RolesOfPSP.componentType = RoleOfPSP()


class PSD2QcType(univ.Sequence):
    pass


PSD2QcType.componentType = namedtype.NamedTypes(
    namedtype.NamedType('rolesOfPSP', RolesOfPSP()),
    namedtype.NamedType('nCAName', NCAName()),
    namedtype.NamedType('nCAId', NCAId())
)


etsi_psd2_policy = _OID(0, 4, 0, 19495, 3)


etsi_psd2_roles = _OID(0, 4, 0, 19495, 1)


id_etsi_psd2_qcStatement = _OID(0, 4, 0, 19495, 2)


id_psd2_role_psp_ai = _OID(0, 4, 0, 19495, 1, 3)


id_psd2_role_psp_as = _OID(0, 4, 0, 19495, 1, 1)


id_psd2_role_psp_ic = _OID(0, 4, 0, 19495, 1, 4)


id_psd2_role_psp_pi = _OID(0, 4, 0, 19495, 1, 2)


id_psd2_role_psp_unspecified = _OID(0, 4, 0, 19495, 1, 0)


qcp_web_psd2 = _OID(0, 4, 0, 19495, 3, 1)


